/*
 * QEMU backup
 *
 * Copyright (C) 2013 Proxmox Server Solutions
 *
 * Authors:
 *  Dietmar Maurer (dietmar@proxmox.com)
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 *
 */

#include "qemu/osdep.h"

#include "trace.h"
#include "block/block.h"
#include "block/block_int.h"
#include "block/blockjob_int.h"
#include "block/block_backup.h"
#include "qapi/error.h"
#include "qapi/qmp/qerror.h"
#include "qemu/ratelimit.h"
#include "qemu/cutils.h"
#include "sysemu/block-backend.h"
#include "qemu/bitmap.h"
#include "qemu/error-report.h"
#include "qapi/qapi-events-block-core.h"

#include "qemu/coroutine_int.h"

#define BACKUP_CLUSTER_SIZE_DEFAULT (1 << 16)
#define BACKUP_LOOP_DELAY 100000000ULL /* ns */
#define WRITE_NOTIFY_TIMEOUT_NS 180000000000LL /* 180 s */

#define NB_WORKERS 64

typedef struct CowRequest {
    int64_t start_byte;
    int64_t end_byte;
    QLIST_ENTRY(CowRequest) list;
    CoQueue wait_queue; /* coroutines blocked on this request */
} CowRequest;

typedef struct NotifierRequest {
    int64_t start; /* in clusters */
    int64_t end; /* in clusters */
    int nb_wait; /* awaited clusters */
    Coroutine *notif;
    QemuCoSleepState *co_sleep_ns_state;
    QSIMPLEQ_ENTRY(NotifierRequest) list;
} NotifierRequest;

typedef struct BackupBlockJob {
    BlockJob common;
    BlockBackend *target;
    /* bitmap for sync=incremental */
    BdrvDirtyBitmap *sync_bitmap;
    MirrorSyncMode sync_mode;
    BlockdevOnError on_source_error;
    BlockdevOnError on_target_error;
    uint64_t len;
    uint64_t bytes_read;
    int64_t cluster_size;
    bool compress;
    NotifierWithReturn before_write;

    HBitmap *copy_bitmap;
    HBitmapIter linear_hbi;

    int async_status;
    CoQueue free_workers;
    bool delayed;
    int running_workers;
    int paused_workers;

    unsigned long *notif_wait_bitmap;
    QSIMPLEQ_HEAD(, NotifierRequest) notifier_reqs;
    NotifierRequest *current_notif;
    HBitmapIter current_notif_hbi;

    bool target_backed_by_source;
    bool skipping_unallocated;
} BackupBlockJob;

static const BlockJobDriver backup_job_driver;

static inline void set_notif_cur(BackupBlockJob *job, NotifierRequest *nr)
{
    if (nr != NULL) {
        hbitmap_iter_init(&job->current_notif_hbi, job->copy_bitmap, nr->start);
    }
    job->current_notif = nr;
}

/* NULL result means that request is already done */
static NotifierRequest *add_notif_req(BackupBlockJob *job,
                                      int64_t start, int64_t end,
                                      Coroutine *notif)
{
    NotifierRequest *nr;
    int nb_wait = bitmap_count_between(job->notif_wait_bitmap, start, end) +
                  hbitmap_count_between(job->copy_bitmap, start, end);

    if (nb_wait == 0) {
        return NULL;
    }

    nr = g_new0(NotifierRequest, 1);

    nr->start = start;
    nr->end = end;
    nr->notif = notif;
    nr->nb_wait = nb_wait;

    QSIMPLEQ_INSERT_TAIL(&job->notifier_reqs, nr, list);
    if (job->current_notif == NULL) {
        set_notif_cur(job, nr);
    }

    return nr;
}

static void detach_notif_req(BackupBlockJob *job, NotifierRequest *req)
{
    assert(req);

    if (job->current_notif == req) {
        set_notif_cur(job, QSIMPLEQ_NEXT(req, list));
    }

    QSIMPLEQ_REMOVE(&job->notifier_reqs, req, NotifierRequest, list);
}

static int64_t next_notif_cluster(BackupBlockJob *job)
{
    while (job->current_notif != NULL) {
        int64_t cluster = hbitmap_iter_next(&job->current_notif_hbi);
        if (cluster != -1 && cluster < job->current_notif->end) {
            return cluster;
        }

        /* nothing to do for current notifier */
        set_notif_cur(job, QSIMPLEQ_NEXT(job->current_notif, list));
    }

    /* nothing to do for notifiers */
    return -1;
}

static void finish_cluster_for_notif(BackupBlockJob *job, int64_t cluster)
{
    NotifierRequest *req, *next;
    QSIMPLEQ_HEAD(, NotifierRequest) finished_reqs;
    QSIMPLEQ_INIT(&finished_reqs);

    trace_backup_finish_cluster_for_notif(job, cluster);

    clear_bit(cluster, job->notif_wait_bitmap);

    /* handle notifiers */
    QSIMPLEQ_FOREACH_SAFE(req, &job->notifier_reqs, list, next) {
        if (cluster >= req->start && cluster < req->end) {
            assert(req->nb_wait > 0);
            req->nb_wait--;
            if (req->nb_wait == 0) {
                detach_notif_req(job, req);
                QSIMPLEQ_INSERT_TAIL(&finished_reqs, req, list);
            }
        }
    }

    while ((req = QSIMPLEQ_FIRST(&finished_reqs))) {
        QSIMPLEQ_REMOVE(&finished_reqs, req, NotifierRequest, list);
        if (req->co_sleep_ns_state) {
            qemu_co_sleep_wake(req->co_sleep_ns_state);
        }
    }
}

static void release_pending_notifiers(BackupBlockJob *job)
{
    NotifierRequest *req, *next;
    QSIMPLEQ_HEAD(, NotifierRequest) finished_reqs;
    QSIMPLEQ_INIT(&finished_reqs);

    QSIMPLEQ_FOREACH_SAFE(req, &job->notifier_reqs, list, next) {
        req->nb_wait = 0;
        detach_notif_req(job, req);
        QSIMPLEQ_INSERT_TAIL(&finished_reqs, req, list);
    }

    while ((req = QSIMPLEQ_FIRST(&finished_reqs))) {
        QSIMPLEQ_REMOVE(&finished_reqs, req, NotifierRequest, list);
        if (req->co_sleep_ns_state) {
            qemu_co_sleep_wake(req->co_sleep_ns_state);
        }
    }
}

static void backup_report_error(BackupBlockJob *job, bool is_read)
{
    /* Taken from block_job_error_action.
     * Actually, this call is equal to call block_job_error_action, because
     * the action should always be 'report' for backup. But is is safer to send
     * event directly. */
    qapi_event_send_block_job_error(job->common.job.id,
                                    is_read ? IO_OPERATION_TYPE_READ :
                                    IO_OPERATION_TYPE_WRITE,
                                    BLOCK_ERROR_ACTION_REPORT, &error_abort);
}

static int coroutine_fn backup_do_read(BackupBlockJob *job,
                                       int64_t offset, int64_t size,
                                       bool is_write_notifier,
                                       QEMUIOVector *qiov)
{
    int ret;

    ret = blk_co_preadv(job->common.blk, offset, size, qiov,
                        is_write_notifier ? BDRV_REQ_NO_SERIALISING : 0);
    if (ret < 0) {
        trace_backup_do_read_fail(job, offset / job->cluster_size, ret);
        backup_report_error(job, true);
    }

    return ret;
}

static int coroutine_fn backup_do_write(BackupBlockJob *job,
                                        int64_t offset, int64_t size,
                                        QEMUIOVector *qiov)
{
    int ret;
    int write_flags = job->target_backed_by_source ?
            BDRV_REQ_SERIALISING | BDRV_REQ_WRITE_UNCHANGED : 0;
    assert(qiov->niov == 1);

    if (buffer_is_zero(qiov->iov->iov_base, qiov->iov->iov_len)) {
        ret = blk_co_pwrite_zeroes(job->target, offset, size,
                                   write_flags | BDRV_REQ_MAY_UNMAP);
    } else {
        ret = blk_co_pwritev(job->target, offset, size, qiov, write_flags |
                             (job->compress ? BDRV_REQ_WRITE_COMPRESSED : 0));
    }
    if (ret < 0) {
        trace_backup_do_write_fail(job, offset / job->cluster_size, ret);
        backup_report_error(job, false);
    }

    return ret;
}

static void coroutine_fn backup_copy_cluster(BackupBlockJob *job,
        int64_t cluster, bool is_write_notifier)
{
    int ret;
    int64_t n;
    struct iovec iov;
    QEMUIOVector bounce_qiov;
    void *bounce_buffer = blk_blockalign(job->common.blk, job->cluster_size);

    trace_backup_copy_cluster_enter(job, cluster);

    n = MIN(job->cluster_size, job->len - cluster * job->cluster_size);

    iov.iov_base = bounce_buffer;
    iov.iov_len = n;
    qemu_iovec_init_external(&bounce_qiov, &iov, 1);

    job->bytes_read += n;

    ret = backup_do_read(job, cluster * job->cluster_size, n,
                         is_write_notifier, &bounce_qiov);
    if (ret < 0) {
        goto out;
    }

    if (job->sync_mode != MIRROR_SYNC_MODE_NONE) {
        finish_cluster_for_notif(job, cluster);
    }

    ret = backup_do_write(job, cluster * job->cluster_size, n,
                          &bounce_qiov);
    if (ret < 0) {
        goto out;
    }

    if (job->sync_mode == MIRROR_SYNC_MODE_NONE) {
        finish_cluster_for_notif(job, cluster);
    }

    /* Publish progress, guest I/O counts as progress too.  Note that the
     * offset field is an opaque progress value, it is not a disk offset.
     */
    job_progress_update(&job->common.job, n);
    ret = 0;

out:
    if (ret < 0 && job->async_status == 0) {
        job->async_status = ret;
    }
    qemu_vfree(bounce_buffer);

    trace_backup_copy_cluster_return(job, cluster, ret);
}

static int coroutine_fn backup_before_write_notify(
        NotifierWithReturn *notifier,
        void *opaque)
{
    BackupBlockJob *job = container_of(notifier, BackupBlockJob, before_write);
    BdrvTrackedRequest *tr = opaque;
    NotifierRequest *nr;
    int64_t start = tr->offset / job->cluster_size;
    int64_t end = DIV_ROUND_UP(tr->offset + tr->bytes, job->cluster_size);

    if ((tr->flags & BDRV_REQ_WRITE_UNCHANGED) &&
        (tr->flags & BDRV_REQ_SERIALISING))
    {
        return 0;
    }

    assert((tr->offset & (BDRV_SECTOR_SIZE - 1)) == 0);
    assert((tr->bytes & (BDRV_SECTOR_SIZE - 1)) == 0);

    nr = add_notif_req(job, start, end, qemu_coroutine_self());

    if (nr == NULL) {
        trace_backup_before_write_notify_skip(job, tr->offset, tr->bytes);
    } else {
        trace_backup_before_write_notify_start(job, tr->offset, tr->bytes, nr,
                                               nr->start, nr->end, nr->nb_wait);

        qemu_co_queue_restart_all(&job->free_workers);
        qemu_co_sleep_ns_wakeable(QEMU_CLOCK_REALTIME, WRITE_NOTIFY_TIMEOUT_NS,
                                  &nr->co_sleep_ns_state);
        if (nr->nb_wait > 0) {
            /* timer expired and read request not finished */
            if (job->async_status == 0) {
                job->async_status = -ETIME;
            }
            detach_notif_req(job, nr);
            trace_backup_before_write_notify_timeout(job, nr, nr->start,
                                                     nr->end, nr->nb_wait);
        } else {
            trace_backup_before_write_notify_success(job, nr, nr->start,
                                                     nr->end, nr->nb_wait);
        }
        g_free(nr);
    }

    return 0;
}

static void backup_cleanup_sync_bitmap(BackupBlockJob *job, int ret)
{
    BdrvDirtyBitmap *bm;
    BlockDriverState *bs = blk_bs(job->common.blk);

    if (ret < 0) {
        /* Merge the successor back into the parent, delete nothing. */
        bm = bdrv_reclaim_dirty_bitmap(bs, job->sync_bitmap, NULL);
        assert(bm);
    } else {
        /* Everything is fine, delete this bitmap and install the backup. */
        bm = bdrv_dirty_bitmap_abdicate(bs, job->sync_bitmap, NULL);
        assert(bm);
    }
}

static void backup_commit(Job *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common.job);
    if (s->sync_bitmap) {
        backup_cleanup_sync_bitmap(s, 0);
    }
}

static void backup_abort(Job *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common.job);
    if (s->sync_bitmap) {
        backup_cleanup_sync_bitmap(s, -1);
    }
}

static void backup_clean(Job *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common.job);
    assert(s->target);
    blk_unref(s->target);
    s->target = NULL;
}

static void backup_attached_aio_context(BlockJob *job, AioContext *aio_context)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common);

    blk_set_aio_context(s->target, aio_context);
}

void backup_do_checkpoint(BlockJob *job, Error **errp)
{
    BackupBlockJob *backup_job = container_of(job, BackupBlockJob, common);
    int64_t len;

    assert(block_job_driver(job) == &backup_job_driver);

    if (backup_job->sync_mode != MIRROR_SYNC_MODE_NONE) {
        error_setg(errp, "The backup job only supports block checkpoint in"
                   " sync=none mode");
        return;
    }

    len = DIV_ROUND_UP(backup_job->len, backup_job->cluster_size);
    hbitmap_set(backup_job->copy_bitmap, 0, len);
}

static void backup_drain(BlockJob *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common);

    /* Need to keep a reference in case blk_drain triggers execution
     * of backup_complete...
     */
    if (s->target) {
        BlockBackend *target = s->target;
        blk_ref(target);
        blk_drain(target);
        blk_unref(target);
    }
}

static void coroutine_fn backup_sleep(BackupBlockJob *job)
{
    uint64_t delay_ns;

    /* We need to yield even for delay_ns = 0 so that bdrv_drain_all() can
     * return. Without a yield, the VM would not reboot. */
    delay_ns = block_job_ratelimit_get_delay(&job->common, job->bytes_read);
    job->bytes_read = 0;
    if (delay_ns > 0) {
        job->delayed = true;
    }
    job_sleep_ns(&job->common.job, delay_ns);
    job->delayed = false;
}

static bool coroutine_fn yield_and_check(BackupBlockJob *job)
{
    if (job_is_cancelled(&job->common.job)) {
        return true;
    }

    backup_sleep(job);

    if (job_is_cancelled(&job->common.job)) {
        return true;
    }

    return false;
}

/* init copy_bitmap from sync_bitmap */
static void backup_incremental_init_copy_bitmap(BackupBlockJob *job)
{
    BdrvDirtyBitmapIter *dbi;
    int64_t offset;
    int64_t end = DIV_ROUND_UP(bdrv_dirty_bitmap_size(job->sync_bitmap),
                               job->cluster_size);

    dbi = bdrv_dirty_iter_new(job->sync_bitmap);
    while ((offset = bdrv_dirty_iter_next(dbi)) != -1) {
        int64_t cluster = offset / job->cluster_size;
        int64_t next_cluster;

        offset += bdrv_dirty_bitmap_granularity(job->sync_bitmap);
        if (offset >= bdrv_dirty_bitmap_size(job->sync_bitmap)) {
            hbitmap_set(job->copy_bitmap, cluster, end - cluster);
            break;
        }

        offset = bdrv_dirty_bitmap_next_zero(job->sync_bitmap, offset,
                                             UINT64_MAX);
        if (offset == -1) {
            hbitmap_set(job->copy_bitmap, cluster, end - cluster);
            break;
        }

        next_cluster = DIV_ROUND_UP(offset, job->cluster_size);
        hbitmap_set(job->copy_bitmap, cluster, next_cluster - cluster);
        if (next_cluster >= end) {
            break;
        }

        bdrv_set_dirty_iter(dbi, next_cluster * job->cluster_size);
    }

    /* TODO job_progress_set_remaining() would make more sense */
    job_progress_update(&job->common.job,
        job->len - hbitmap_count(job->copy_bitmap) * job->cluster_size);

    bdrv_dirty_iter_free(dbi);
}

static void backup_skip_clusters(BackupBlockJob *job,
                                 int64_t start,
                                 int64_t end)
{
    int64_t n_bytes;

    if (end * job->cluster_size > job->len) {
        end--;
        n_bytes = job->len - end * job->cluster_size;
        assert(n_bytes > 0);

        if (hbitmap_get(job->copy_bitmap, end)) {
            trace_backup_skip_cluster(job, end);
            hbitmap_reset(job->copy_bitmap, end, 1);
            job_progress_update(&job->common.job, n_bytes);
            finish_cluster_for_notif(job, end);
        }
    }

    for ( ; start < end; start++) {
        if (!hbitmap_get(job->copy_bitmap, start)) {
            continue;
        }

        trace_backup_skip_cluster(job, start);
        hbitmap_reset(job->copy_bitmap, start, 1);
        job_progress_update(&job->common.job, job->cluster_size);
        finish_cluster_for_notif(job, start);
    }
}

static inline int64_t max_query_bytes(BackupBlockJob *job)
{
    return INT_MAX & ~(job->cluster_size - 1);
}

static void backup_skip_unallocated_clusters(BackupBlockJob *job,
                                             BlockDriverState *base)
{
    int ret;
    BlockDriverState *bs = blk_bs(job->common.blk);
    int64_t offset = 0;

    while (offset < job->len) {
        int64_t n;
        int64_t nb_bytes = MIN(max_query_bytes(job), job->len - offset);
        int64_t cluster = offset / job->cluster_size;

        if (!hbitmap_get(job->copy_bitmap, cluster)) {
            offset += job->cluster_size;
            continue;
        }

        if (yield_and_check(job) || job->async_status < 0) {
            return;
        }

        ret = bdrv_is_allocated_above(bs, base, offset, nb_bytes, &n);
        if (job->async_status < 0) {
            return;
        }
        if (ret < 0) {
            job->async_status = ret;
            return;
        }

        if (ret == 1 || n < job->cluster_size) {
            offset += DIV_ROUND_UP(n, job->cluster_size) * job->cluster_size;
            continue;
        }

        if (offset + n < job->len) {
            n -= n % job->cluster_size;
        }

        backup_skip_clusters(job, cluster,
                DIV_ROUND_UP(offset + n, job->cluster_size));

        assert(n > 0);
        offset += n;
    }
}

#define BACKUP_WORKER_STOP -1
#define BACKUP_WORKER_PAUSE -2

static inline int64_t backup_get_work(BackupBlockJob *job,
                                      bool *is_write_notifier)
{
    int64_t cluster;

    assert(is_write_notifier != NULL);

    if (job_is_cancelled(&job->common.job) || job->async_status != 0) {
        return BACKUP_WORKER_STOP;
    }

    cluster = next_notif_cluster(job);
    *is_write_notifier = cluster != -1;
    if (cluster == -1) {
        uint64_t delay_ns = 0;

        if (job->common.speed) {
            delay_ns = block_job_ratelimit_get_delay(&job->common,
                                                     job->bytes_read);
            job->bytes_read = 0;
        }

        if (job_should_pause(&job->common.job) || job->delayed ||
            job->sync_mode == MIRROR_SYNC_MODE_NONE || (delay_ns > 0) ||
            job->skipping_unallocated)
        {
            return BACKUP_WORKER_PAUSE;
        }

        cluster = hbitmap_iter_next(&job->linear_hbi);
        if (cluster == -1) {
            return BACKUP_WORKER_STOP;
        }
    }

    assert(hbitmap_get(job->copy_bitmap, cluster));
    hbitmap_reset(job->copy_bitmap, cluster, 1);
    set_bit(cluster, job->notif_wait_bitmap);

    return cluster;
}

static void backup_worker_pause(BackupBlockJob *job)
{
    job->paused_workers++;
    qemu_co_queue_wait(&job->free_workers, NULL);
    job->paused_workers--;
}

static void coroutine_fn backup_worker_co(void *opaque)
{
    BackupBlockJob *job = opaque;

    job->running_workers++;

    while (true) {
        bool is_write_notifier = false;
        int64_t cluster = backup_get_work(job, &is_write_notifier);

        trace_backup_worker_co(job, cluster, is_write_notifier);

        switch (cluster) {
        case BACKUP_WORKER_STOP:
            job->running_workers--;
            return;
        case BACKUP_WORKER_PAUSE:
            backup_worker_pause(job);
            break;
        default:
            backup_copy_cluster(job, cluster, is_write_notifier);
        }
    }
}

/* this loop runs in block job coroutine (job->common.co) and handles
 * block-job specific pauses, delays and yields. It doesn't do real data copy.
 */
static void coroutine_fn backup_loop(BackupBlockJob *job)
{
    while (job->running_workers > 0) {
        backup_sleep(job);

        /* we may be resumed after pause or delay, so we should resume paused
         * workers */
        qemu_co_queue_restart_all(&job->free_workers);

        qemu_co_sleep_ns(QEMU_CLOCK_REALTIME, BACKUP_LOOP_DELAY);
    }
}

static void coroutine_fn backup_pause(Job *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common.job);

    while (s->paused_workers < s->running_workers && job_should_pause(job)) {
        qemu_co_sleep_ns(QEMU_CLOCK_REALTIME, BACKUP_LOOP_DELAY);
    }
}

static int coroutine_fn backup_run(Job *job, Error **errp)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common.job);
    BlockDriverState *bs = blk_bs(s->common.blk);
    int64_t nb_clusters;
    bool is_top = s->sync_mode == MIRROR_SYNC_MODE_TOP;
    bool is_full = s->sync_mode == MIRROR_SYNC_MODE_FULL;
    int i;

    trace_backup_run_start(s, bs, bdrv_get_device_or_node_name(bs));

    qemu_co_queue_init(&s->free_workers);
    QSIMPLEQ_INIT(&s->notifier_reqs);

    nb_clusters = DIV_ROUND_UP(s->len, s->cluster_size);
    job_progress_set_remaining(job, s->len);

    s->skipping_unallocated = is_top || (is_full
        /* && bdrv_has_zero_init(blk_bs(s->target)) */);
    /*
     * We are sure that in vz-7.6 target for full backup is an empty qcow2
     * image. However we can't check it from qemu (bdrv_has_zero_init do
     * not work for NBD). So, bdrv_has_zero_init() call has been commented up.
     * Otherwise backing up empty (or almost empty) images is too long
     * #PSBM-75655.
     */

    s->notif_wait_bitmap = bitmap_new(nb_clusters);
    s->copy_bitmap = hbitmap_alloc(nb_clusters, 0);
    if (s->sync_mode == MIRROR_SYNC_MODE_INCREMENTAL) {
        backup_incremental_init_copy_bitmap(s);
    } else {
        hbitmap_set(s->copy_bitmap, 0, nb_clusters);
    }
    hbitmap_iter_init(&s->linear_hbi, s->copy_bitmap, 0);

    for (i = 0; i < NB_WORKERS; ++i) {
        Coroutine *co = qemu_coroutine_create(backup_worker_co, job);
        qemu_coroutine_enter(co);
    }

    s->before_write.notify = backup_before_write_notify;
    bdrv_add_before_write_notifier(bs, &s->before_write);

    if (s->sync_mode == MIRROR_SYNC_MODE_NONE) {
        /* All bits are set in copy_bitmap to allow any cluster to be copied.
         * This does not actually require them to be copied. */
        while (!job_is_cancelled(job)) {
            /* Yield until the job is cancelled.  We just let backup workers
             * service CoW requests. */
            job_yield(job);
        }
    }
    if (s->skipping_unallocated) {
        BlockDriverState *base = is_top ? backing_bs(bs) : NULL;
        backup_skip_unallocated_clusters(s, base);
        s->skipping_unallocated = false;
    }

    /* wait for all workers to exit */
    backup_loop(s);

    notifier_with_return_remove(&s->before_write);

    /* in case of error or cancel there may be pending notifiers */
    release_pending_notifiers(s);

    hbitmap_free(s->copy_bitmap);
    g_free(s->notif_wait_bitmap);

    trace_backup_run_end(s, bs, bdrv_get_device_or_node_name(bs),
                         s->async_status);

    return s->async_status;
}

static const BlockJobDriver backup_job_driver = {
    .job_driver = {
        .instance_size          = sizeof(BackupBlockJob),
        .job_type               = JOB_TYPE_BACKUP,
        .free                   = block_job_free,
        .user_resume            = block_job_user_resume,
        .drain                  = block_job_drain,
        .run                    = backup_run,
        .commit                 = backup_commit,
        .abort                  = backup_abort,
        .clean                  = backup_clean,
        .pause                  = backup_pause,
    },
    .attached_aio_context   = backup_attached_aio_context,
    .drain                  = backup_drain,
};

BlockJob *backup_job_create(const char *job_id, BlockDriverState *bs,
                  BlockDriverState *target, int64_t speed,
                  MirrorSyncMode sync_mode, BdrvDirtyBitmap *sync_bitmap,
                  bool compress,
                  BlockdevOnError on_source_error,
                  BlockdevOnError on_target_error,
                  int creation_flags,
                  BlockCompletionFunc *cb, void *opaque,
                  JobTxn *txn, Error **errp)
{
    int64_t len;
    BlockDriverInfo bdi;
    BackupBlockJob *job = NULL;
    int ret;

    assert(bs);
    assert(target);

    if (bs == target) {
        error_setg(errp, "Source and target cannot be the same");
        return NULL;
    }

    /* virtuozzo: to simplify async backup scheme we have dropped
     * retrying of io operations in case of errors.
     */
    if (on_source_error != BLOCKDEV_ON_ERROR_REPORT ||
        on_target_error != BLOCKDEV_ON_ERROR_REPORT) {
        error_setg(errp, "Only 'report' value of on-source-error and "
                         "on-target-error parameters is supported.");
        return NULL;
    }

    if (!bdrv_is_inserted(bs)) {
        error_setg(errp, "Device is not inserted: %s",
                   bdrv_get_device_name(bs));
        return NULL;
    }

    if (!bdrv_is_inserted(target)) {
        error_setg(errp, "Device is not inserted: %s",
                   bdrv_get_device_name(target));
        return NULL;
    }

    if (compress && target->drv->bdrv_co_pwritev_compressed == NULL) {
        error_setg(errp, "Compression is not supported for this drive %s",
                   bdrv_get_device_name(target));
        return NULL;
    }

    if (bdrv_op_is_blocked(bs, BLOCK_OP_TYPE_BACKUP_SOURCE, errp)) {
        return NULL;
    }

    if (bdrv_op_is_blocked(target, BLOCK_OP_TYPE_BACKUP_TARGET, errp)) {
        return NULL;
    }

    if (sync_mode == MIRROR_SYNC_MODE_INCREMENTAL) {
        if (!sync_bitmap) {
            error_setg(errp, "must provide a valid bitmap name for "
                             "\"incremental\" sync mode");
            return NULL;
        }

        /* Create a new bitmap, and freeze/disable this one. */
        if (bdrv_dirty_bitmap_create_successor(bs, sync_bitmap, errp) < 0) {
            return NULL;
        }
    } else if (sync_bitmap) {
        error_setg(errp,
                   "a sync_bitmap was provided to backup_run, "
                   "but received an incompatible sync_mode (%s)",
                   MirrorSyncMode_str(sync_mode));
        return NULL;
    }

    len = bdrv_getlength(bs);
    if (len < 0) {
        error_setg_errno(errp, -len, "unable to get length for '%s'",
                         bdrv_get_device_name(bs));
        goto error;
    }

    /* job->len is fixed, so we can't allow resize */
    job = block_job_create(job_id, &backup_job_driver, txn, bs,
                           BLK_PERM_CONSISTENT_READ,
                           BLK_PERM_CONSISTENT_READ | BLK_PERM_WRITE |
                           BLK_PERM_WRITE_UNCHANGED | BLK_PERM_GRAPH_MOD,
                           speed, creation_flags, cb, opaque, errp);
    if (!job) {
        goto error;
    }

    /* Detect image-fleecing (and similar) schemes */
    job->target_backed_by_source = bdrv_chain_contains(target, bs);

    /* The target must match the source in size, so no resize here either */
    job->target = blk_new(job->target_backed_by_source ?
                              BLK_PERM_WRITE_UNCHANGED : BLK_PERM_WRITE,
                          BLK_PERM_CONSISTENT_READ | BLK_PERM_WRITE |
                          BLK_PERM_WRITE_UNCHANGED | BLK_PERM_GRAPH_MOD);
    ret = blk_insert_bs(job->target, target, errp);
    if (ret < 0) {
        goto error;
    }

    job->on_source_error = on_source_error;
    job->on_target_error = on_target_error;
    job->sync_mode = sync_mode;
    job->sync_bitmap = sync_mode == MIRROR_SYNC_MODE_INCREMENTAL ?
                       sync_bitmap : NULL;
    job->compress = compress;

    /* If there is no backing file on the target, we cannot rely on COW if our
     * backup cluster size is smaller than the target cluster size. Even for
     * targets with a backing file, try to avoid COW if possible. */
    ret = bdrv_get_info(target, &bdi);
    if (ret == -ENOTSUP && !target->backing) {
        /* Cluster size is not defined */
        warn_report("The target block device doesn't provide "
                    "information about the block size and it doesn't have a "
                    "backing file. The default block size of %u bytes is "
                    "used. If the actual block size of the target exceeds "
                    "this default, the backup may be unusable",
                    BACKUP_CLUSTER_SIZE_DEFAULT);
        job->cluster_size = BACKUP_CLUSTER_SIZE_DEFAULT;
    } else if (ret < 0 && !target->backing) {
        error_setg_errno(errp, -ret,
            "Couldn't determine the cluster size of the target image, "
            "which has no backing file");
        error_append_hint(errp,
            "Aborting, since this may create an unusable destination image\n");
        goto error;
    } else if (ret < 0 && target->backing) {
        /* Not fatal; just trudge on ahead. */
        job->cluster_size = BACKUP_CLUSTER_SIZE_DEFAULT;
    } else {
        job->cluster_size = MAX(BACKUP_CLUSTER_SIZE_DEFAULT, bdi.cluster_size);
        if (job->cluster_size != bdi.cluster_size) {
            warn_report("Backup job cluster size %" PRId64 " doesn't match the "
                        "one of target image '%s': %d, it may be inefficient.",
                        job->cluster_size, target->filename, bdi.cluster_size);
        }
        /* PSBM-104879 */
        if (bdi.cluster_size != 65536) {
            warn_report("The cluster size of %d for the image '%s' is not "
                        "equal to 64K.", bdi.cluster_size, target->filename);
        }
    }

    if (job->sync_bitmap) {
        uint32_t granularity = bdrv_dirty_bitmap_granularity(job->sync_bitmap);
        if (job->cluster_size != granularity) {
            warn_report("Backup job cluster size %" PRId64 " doesn't match the "
                        "granularity %" PRIu32 " of the chosen bitmap '%s', "
                        "it may be inefficient.", job->cluster_size,
                        granularity, bdrv_dirty_bitmap_name(job->sync_bitmap));
        }
    }

    /* Required permissions are already taken with target's blk_new() */
    block_job_add_bdrv(&job->common, "target", target, 0, BLK_PERM_ALL,
                       &error_abort);
    if (job->target_backed_by_source) {
        /* image-fleecing like scheme: allow another backup from target */
        bdrv_op_unblock(target, BLOCK_OP_TYPE_BACKUP_SOURCE,
                        job->common.blocker);
    }
    job->len = len;

    return &job->common;

 error:
    if (sync_bitmap) {
        bdrv_reclaim_dirty_bitmap(bs, sync_bitmap, NULL);
    }
    if (job) {
        backup_clean(&job->common.job);
        job_early_fail(&job->common.job);
    }

    return NULL;
}
