/*
 * QEMU Hyper-V VSS
 *
 * Copyright (c) 2018 Virtuozzo International GmbH.
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 */

#include "qemu/osdep.h"
#include "qemu/error-report.h"
#include "qemu/iov.h"
#include "qapi/error.h"
#include "qapi/qapi-events-misc.h"
#include "qapi/qapi-commands-misc.h"
#include "hw/vmbus/vmbus.h"
#include "hw/scsi/scsi.h"
#include "hv-util-proto.h"
#include "block/block.h"
#include "sysemu/block-backend.h"

enum vss_op {
    VSS_OP_CREATE = 0,
    VSS_OP_DELETE,
    VSS_OP_HOT_BACKUP,
    VSS_OP_GET_DM_INFO,
    VSS_OP_BU_COMPLETE,
    /*
     * Following operations are only supported with IC version >= 5.0
     */
    VSS_OP_FREEZE, /* Freeze the file systems in the VM */
    VSS_OP_THAW, /* Unfreeze the file systems */
    VSS_OP_AUTO_RECOVER,
    VSS_OP_COUNT /* Number of operations, must be last */
};

struct vss_hdr {
    uint8_t operation;
    uint8_t reserved[7];
} QEMU_PACKED;

struct vss_hot_backup {
    uint8_t reserved[24];
} QEMU_PACKED;

struct vss_freeze_drive {
#define VSS_FREEZE_DRIVE_SCSI 1
#define VSS_FREEZE_DRIVE_IDE 2
        uint32_t type;
        uint8_t device_guid[16];
        uint8_t channel;
        uint8_t id;
        uint16_t lun;
} QEMU_PACKED;

struct vss_freeze {
    uint32_t flags;
    uint32_t reserved;
    uint32_t drive_count;
/* Most likely, it corresponds to 256 SCSI drives plus 4 IDE drives */
#define VSS_FREEZE_DRIVE_MAX 260
    struct vss_freeze_drive drive[VSS_FREEZE_DRIVE_MAX];
} QEMU_PACKED;

struct vss_thaw {
    uint8_t reserved[24];
} QEMU_PACKED;

static const struct ic_version hv_vss_versions[] = {
    { 5, 0 },
    { 6, 0 },
};

#define TYPE_HV_VSS "hv-vss"
#define HV_VSS(obj) OBJECT_CHECK(HvVss, (obj), TYPE_HV_VSS)
#define HV_VSS_GUID "35fa2e29-ea23-4236-96ae-3a6ebacba440"
#define HV_VSS_INSTANCEID "2450ee40-33bf-4fbd-892e-9fb06e9214cf"

typedef enum HvVssStatus HvVssStatus;
enum HvVssStatus {
    HV_VSS_STATUS_IDLE,
    HV_VSS_STATUS_SEND,
    HV_VSS_STATUS_RECV,
};

typedef enum HvVssCommand HvVssCommand;
enum HvVssCommand {
    HV_VSS_COMMAND_NEGOTIATE,
    HV_VSS_COMMAND_FREEZE,
    HV_VSS_COMMAND_THAW,
};

typedef enum HvVssResult HvVssResult;
enum HvVssResult {
    HV_VSS_RESULT_OK,
    HV_VSS_RESULT_FAIL,
    HV_VSS_RESULT_BROKEN,
};

typedef struct HvVss HvVss;
struct HvVss {
    VMBusDevice parent;

    HvVssCommand command;
    HvVssStatus status;
    HvVssResult result;

    bool thaw_on_freeze;
    bool freeze_hot_backup;
    uint32_t freeze_drive_count;
    struct vss_freeze_drive freeze_drive[VSS_FREEZE_DRIVE_MAX];

    uint8_t transaction;
    struct ic_version version;
};

static void hv_vss_vmdev_reset(VMBusDevice *vdev)
{
    HvVss *vss = HV_VSS(vdev);

    vss->transaction = (uint8_t)(-1);

    vss->command = HV_VSS_COMMAND_NEGOTIATE;
    vss->status = HV_VSS_STATUS_SEND;
    vss->result = HV_VSS_RESULT_OK;
}

static void hv_vss_vmdev_realize(VMBusDevice *vdev, Error **errp)
{
    hv_vss_vmdev_reset(vdev);
}

static int hv_vss_msg_send(VMBusChannel *chan, uint16_t type, void *buf,
                           uint32_t icmsgsize)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);
    struct vmbuspipe_hdr *vmbuspipe = buf;
    struct icmsg_hdr *icmsg = (void *)(vmbuspipe + 1);
    uint32_t total;
    int ret;

    icmsg->icmsgtype = type;
    if (icmsg->icmsgtype != ICMSGTYPE_NEGOTIATE) {
        icmsg->icverframe.major = HV_UTIL_FW_MAJOR;
        icmsg->icverframe.minor = HV_UTIL_FW_MINOR;
        icmsg->icvermsg = vss->version;
    }
    icmsg->icflags = ICMSGHDRFLAG_TRANSACTION | ICMSGHDRFLAG_REQUEST;
    icmsg->icmsgsize = icmsgsize;

    vss->transaction++;
    icmsg->ictransaction_id = vss->transaction;

    vmbuspipe->flags = 0x00000001; /* Magic value */
    vmbuspipe->msgsize = icmsg->icmsgsize + sizeof(struct icmsg_hdr);

    total = vmbuspipe->msgsize + sizeof(struct vmbuspipe_hdr);

    ret = vmbus_channel_reserve(chan, 0, total);
    if (ret) {
        return -EAGAIN;
    }

    ret = vmbus_channel_send(chan, VMBUS_PACKET_DATA_INBAND, NULL, 0, vmbuspipe,
                             total, false, 0);
    return 0;
}

static void hv_vss_negotiate_send(VMBusChannel *chan)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);
    int ret;

    struct {
        struct vmbuspipe_hdr vmbuspipe;
        struct icmsg_hdr icmsg;
        struct icmsg_negotiate nego;
        struct ic_version icframe_ver;
        struct ic_version icmsg_ver[ARRAY_SIZE(hv_vss_versions)];
    } QEMU_PACKED buf = { };
    uint32_t icmsgsize;

    buf.nego.icframe_vercnt = 1;
    buf.icframe_ver.major = HV_UTIL_FW_MAJOR;
    buf.icframe_ver.minor = HV_UTIL_FW_MINOR;

    buf.nego.icmsg_vercnt = ARRAY_SIZE(hv_vss_versions);
    memcpy(buf.icmsg_ver, hv_vss_versions, sizeof(hv_vss_versions));

    icmsgsize = offsetof(struct icmsg_negotiate,
                         icversion_data[buf.nego.icframe_vercnt +
                                        buf.nego.icmsg_vercnt]);
    ret = hv_vss_msg_send(chan, ICMSGTYPE_NEGOTIATE, &buf, icmsgsize);
    if (ret == -EAGAIN) {
        return;
    }
    if (ret) {
        vss->status = HV_VSS_STATUS_IDLE;
        vss->result = HV_VSS_RESULT_BROKEN;
        return;
    }

    vss->status = HV_VSS_STATUS_RECV;
}

static void hv_vss_negotiate_recv(VMBusChannel *chan, void *msg,
                                  uint32_t msgsize)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);
    bool matched;
    int i, j;

    struct icmsg_negotiate *nego;
    struct ic_version *ver;

    nego = msg;
    if (msg + msgsize < (void *)(nego + 1)) {
        vss->result = HV_VSS_RESULT_BROKEN;
        goto complete;
    }

    if (msgsize < offsetof(struct icmsg_negotiate,
                           icversion_data[nego->icframe_vercnt +
                                          nego->icmsg_vercnt])) {
        vss->result = HV_VSS_RESULT_BROKEN;
        goto complete;
    }

    matched = false;
    for (i = 0; i < nego->icframe_vercnt; i++) {
        ver = nego->icversion_data + i;
        if (ver->major == HV_UTIL_FW_MAJOR &&
                ver->minor == HV_UTIL_FW_MINOR) {
            matched = true;
            break;
        }
    }
    if (!matched) {
        vss->result = HV_VSS_RESULT_FAIL;
        goto complete;
    }

    matched = false;
    for (i = 0; i < nego->icmsg_vercnt; i++) {
        ver = nego->icversion_data + nego->icframe_vercnt + i;
        for (j = 0; ARRAY_SIZE(hv_vss_versions) && !matched; j++) {
            if (ver->major == hv_vss_versions[j].major &&
                    ver->minor == hv_vss_versions[j].minor) {
                vss->version = hv_vss_versions[j];
                matched = true;
                break;
            }
        }
        if (matched) {
            break;
        }
    }
    if (!matched) {
        vss->result = HV_VSS_RESULT_FAIL;
        goto complete;
    }

    vss->result = HV_VSS_RESULT_OK;

complete:
    vss->status = HV_VSS_STATUS_IDLE;
}

static void hv_vss_op_send(VMBusChannel *chan)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);
    int ret;

    struct {
        struct vmbuspipe_hdr vmbuspipe;
        struct icmsg_hdr icmsg;
        struct vss_hdr op;
        union {
            struct vss_hot_backup hot_backup;
            struct vss_freeze freeze;
            struct vss_thaw thaw;
        } data;
    } QEMU_PACKED buf = { };
    uint32_t icmsgsize = sizeof(struct vss_hdr);

    switch (vss->command) {
    case HV_VSS_COMMAND_FREEZE:
        if (vss->freeze_hot_backup) {
            buf.op.operation = VSS_OP_HOT_BACKUP;
            icmsgsize += sizeof(struct vss_hot_backup);
        } else {
            memcpy(&buf.data.freeze.drive, vss->freeze_drive,
                   sizeof(struct vss_freeze_drive) * vss->freeze_drive_count);
            buf.data.freeze.drive_count = vss->freeze_drive_count;
            buf.data.freeze.flags = 5;
            buf.op.operation = VSS_OP_FREEZE;
            icmsgsize += sizeof(struct vss_freeze);
        }
        break;
    case HV_VSS_COMMAND_THAW:
        buf.op.operation = VSS_OP_THAW;
        icmsgsize += sizeof(struct vss_thaw);
        break;
    default:
        assert(false);
        break;
    }

    ret = hv_vss_msg_send(chan, ICMSGTYPE_VSS, &buf, icmsgsize);
    if (ret == -EAGAIN) {
        return;
    }
    if (ret) {
        vss->status = HV_VSS_STATUS_IDLE;
        vss->result = HV_VSS_RESULT_BROKEN;
        return;
    }

    vss->status = HV_VSS_STATUS_RECV;
}

static void hv_vss_freeze_recv(VMBusChannel *chan, void *msg, uint32_t msgsize)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);

    if (vss->freeze_hot_backup) {
        vss->freeze_hot_backup = false;
        vss->status = HV_VSS_STATUS_SEND;
    } else {
        vss->status = HV_VSS_STATUS_IDLE;
        vss->result = HV_VSS_RESULT_OK;
    }
}

static void hv_vss_thaw_recv(VMBusChannel *chan, void *msg, uint32_t msgsize)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);

    vss->status = HV_VSS_STATUS_IDLE;
    vss->result = HV_VSS_RESULT_OK;
}

static void hv_vss_cmd_send(VMBusChannel *chan)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);

    switch (vss->command) {
    case HV_VSS_COMMAND_NEGOTIATE:
        hv_vss_negotiate_send(chan);
        break;
    case HV_VSS_COMMAND_FREEZE:
    case HV_VSS_COMMAND_THAW:
        hv_vss_op_send(chan);
        break;
    }
}

static const char *hv_vss_command_name(HvVssCommand command)
{
    switch (command) {
    case HV_VSS_COMMAND_NEGOTIATE:
        return "negotiate";
    case HV_VSS_COMMAND_FREEZE:
        return "freeze";
    case HV_VSS_COMMAND_THAW:
        return "thaw";
    }
    return "";
}

static const char *hv_vss_result_name(HvVssResult result)
{
    switch (result) {
    case HV_VSS_RESULT_OK:
        return "ok";
    case HV_VSS_RESULT_FAIL:
        return "fail";
    case HV_VSS_RESULT_BROKEN:
        return "broken";
    }
    return "";
}

static void hv_vss_cmd_event(VMBusChannel *chan)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);

    error_report("info: hv-vss: command=%s result=%s",
                 hv_vss_command_name(vss->command),
                 hv_vss_result_name(vss->result));

    if (vss->result == HV_VSS_RESULT_BROKEN) {
        qapi_event_send_hv_vss_broken(&error_abort);
        return;
    }

    switch (vss->command) {
    case HV_VSS_COMMAND_NEGOTIATE:
        if (vss->result == HV_VSS_RESULT_OK) {
            qapi_event_send_hv_vss_thawed(false, &error_abort);
        } else {
            qapi_event_send_hv_vss_broken(&error_abort);
        }
        break;
    case HV_VSS_COMMAND_FREEZE:
        if (vss->result == HV_VSS_RESULT_OK) {
            qapi_event_send_hv_vss_frozen(&error_abort);
        } else {
            qapi_event_send_hv_vss_thawed(false, &error_abort);
        }
        break;
    case HV_VSS_COMMAND_THAW:
        if (vss->result == HV_VSS_RESULT_OK) {
            qapi_event_send_hv_vss_thawed(!vss->thaw_on_freeze, &error_abort);
        } else {
            qapi_event_send_hv_vss_thawed(false, &error_abort);
        }
        break;
    }
}

static void hv_vss_cmd_recv(VMBusChannel *chan)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);
    VMBusChanReq *req;

    struct vmbuspipe_hdr *vmbuspipe;
    struct icmsg_hdr *icmsg;
    void *msg;

    req = vmbus_channel_recv(chan, sizeof(VMBusChanReq));
    if (!req) {
        goto out;
    }

    vmbuspipe = req->msg;
    if (req->msg + req->msglen < (void *)(vmbuspipe + 1)) {
        vss->status = HV_VSS_STATUS_IDLE;
        vss->result = HV_VSS_RESULT_BROKEN;
        goto out;
    }

    icmsg = (void *)(vmbuspipe + 1);
    if (req->msg + req->msglen < (void *)(icmsg + 1)) {
        vss->status = HV_VSS_STATUS_IDLE;
        vss->result = HV_VSS_RESULT_BROKEN;
        goto out;
    }

    if (icmsg->ictransaction_id != vss->transaction) {
        goto out;
    }

    if (icmsg->status) {
        vss->status = HV_VSS_STATUS_IDLE;
        vss->result = HV_VSS_RESULT_FAIL;
        goto out;
    }

    msg = (void *)(icmsg + 1);
    if (req->msg + req->msglen < msg + icmsg->icmsgsize) {
        vss->status = HV_VSS_STATUS_IDLE;
        vss->result = HV_VSS_RESULT_BROKEN;
        goto out;
    }

    switch (vss->command) {
    case HV_VSS_COMMAND_NEGOTIATE:
        hv_vss_negotiate_recv(chan, msg, icmsg->icmsgsize);
        break;
    case HV_VSS_COMMAND_FREEZE:
        hv_vss_freeze_recv(chan, msg, icmsg->icmsgsize);
        break;
    case HV_VSS_COMMAND_THAW:
        hv_vss_thaw_recv(chan, msg, icmsg->icmsgsize);
        break;
    }

out:
    vmbus_release_req(req);

    if (vss->status == HV_VSS_STATUS_IDLE) {
        hv_vss_cmd_event(chan);
    }
}

static void hv_vss_notify_cb(VMBusChannel *chan)
{
    VMBusDevice *vdev = vmbus_channel_device(chan);
    HvVss *vss = HV_VSS(vdev);
    HvVssStatus old_status;

    do {
        old_status = vss->status;

        switch (vss->status) {
        case HV_VSS_STATUS_IDLE:
            break;
        case HV_VSS_STATUS_SEND:
            hv_vss_cmd_send(chan);
            break;
        case HV_VSS_STATUS_RECV:
            hv_vss_cmd_recv(chan);
            break;
        }
    } while (old_status != vss->status);
}

static HvVss *hv_vss_instance(void)
{
    return HV_VSS(object_resolve_path_type("", TYPE_HV_VSS, NULL));
}

static int hv_vss_get_drive(DeviceState *dev, struct vss_freeze_drive *drive)
{
    QemuUUID instanceid;
    VMBusDevice *vdev;
    SCSIDevice *sdev;
    BusState *bus;

    if (!object_dynamic_cast(OBJECT(dev), TYPE_SCSI_DEVICE)) {
        return -ENODEV;
    }

    sdev = SCSI_DEVICE(dev);
    bus = qdev_get_parent_bus(dev);
    assert(object_dynamic_cast(OBJECT(bus), TYPE_SCSI_BUS));
    assert(bus->parent);

    if (!object_dynamic_cast(OBJECT(bus->parent), TYPE_VMBUS_DEVICE)) {
        return -ENODEV;
    }

    vdev = VMBUS_DEVICE(bus->parent);
    instanceid = vdev->instanceid;
    qemu_uuid_bswap(&instanceid);

    memset(drive, 0, sizeof(struct vss_freeze_drive));

    drive->type = VSS_FREEZE_DRIVE_SCSI;
    memcpy(drive->device_guid, instanceid.data, 16);

    drive->channel = sdev->channel;
    drive->id = sdev->id;
    drive->lun = sdev->lun;

    return 0;
}

typedef struct HvVssWalker {
    struct vss_freeze_drive *drive;
    uint32_t count;
    Error **errp;
} HvVssWalker;

static int qmp_hv_vss_freeze_drive_walkerfn(DeviceState *dev, void *opaque)
{
    HvVssWalker *walker = opaque;
    struct vss_freeze_drive drive;
    int ret;

    ret = hv_vss_get_drive(dev, &drive);
    if (ret) {
        return 0;
    }

    if (walker->count == VSS_FREEZE_DRIVE_MAX) {
        error_setg(walker->errp, "Too many drives");
        return -ENODEV;
    }

    walker->drive[walker->count] = drive;
    walker->count++;
    return 0;
}

static uint32_t qmp_hv_vss_freeze_drive_all(struct vss_freeze_drive *drive,
                                            Error **errp)
{
    HvVssWalker walker = {
        .drive = drive,
        .count = 0,
        .errp = errp
    };

    qbus_walk_children(sysbus_get_default(), qmp_hv_vss_freeze_drive_walkerfn,
                       NULL, NULL, NULL, &walker);
    return walker.count;
}

static uint32_t qmp_hv_vss_freeze_drive_list(struct vss_freeze_drive *drive,
                                             strList *ids, Error **errp)
{
    uint32_t count = 0;
    strList *id;
    int ret;

    for (id = ids; id; id = id->next) {
        BlockBackend *blk;
        DeviceState *dev;

        if (count == VSS_FREEZE_DRIVE_MAX) {
            error_setg(errp, "Too many drives");
            goto fail;
        }

        blk = blk_by_name(id->value);
        if (!blk) {
            error_setg(errp, "Drive with id '%s' doesn't exist", id->value);
            goto fail;
        }

        dev = blk_get_attached_dev(blk);
        if (!dev) {
            error_setg(errp, "Drive with id '%s' doesn't have attachement",
                       id->value);
            goto fail;
        }

        ret = hv_vss_get_drive(dev, drive);
        if (ret) {
            error_setg(errp, "Drive with id '%s' is not VSS-capable",
                       id->value);
            goto fail;
        }

        count++;
        drive++;
    }

    return count;

fail:
    return 0;
}

static void qmp_hv_vss_cmd(bool freeze, void *arg, Error **errp)
{
    HvVss *vss = hv_vss_instance();
    VMBusChannel *chan;

    if (!vss) {
        error_setg(errp, "No VSS");
        return;
    }

    switch (vss->status) {
    case HV_VSS_STATUS_IDLE:
        break;
    default:
        if (!freeze && vss->command == HV_VSS_COMMAND_FREEZE) {
            /* Permit `thaw' command during `freeze' processing */
            break;
        }
        error_setg(errp, "VSS is busy");
        return;
    }

    if (vss->result == HV_VSS_RESULT_BROKEN ||
            (vss->result == HV_VSS_RESULT_FAIL &&
             vss->command == HV_VSS_COMMAND_NEGOTIATE)) {
        error_setg(errp, "VSS is broken");
        return;
    }

    if (freeze) {
        strList *ids = arg;
        uint32_t count;

        if (ids) {
            count = qmp_hv_vss_freeze_drive_list(vss->freeze_drive, ids, errp);
        } else {
            count = qmp_hv_vss_freeze_drive_all(vss->freeze_drive, errp);
        }
        if (*errp) {
            return;
        }
        if (!count) {
            error_setg(errp, "No VSS-capable drives");
            return;
        }

        vss->freeze_drive_count = count;
        vss->freeze_hot_backup = true;
        vss->command = HV_VSS_COMMAND_FREEZE;
    } else {
        vss->thaw_on_freeze = vss->status != HV_VSS_STATUS_IDLE;
        vss->command = HV_VSS_COMMAND_THAW;
    }

    vss->status = HV_VSS_STATUS_SEND;

    chan = vmbus_device_channel(VMBUS_DEVICE(vss), 0);
    vmbus_notify_channel(chan);
}

void qmp_hv_vss_freeze(bool has_ids, strList *ids, Error **errp)
{
    qmp_hv_vss_cmd(true, (has_ids && ids ? ids : NULL), errp);
}

void qmp_hv_vss_thaw(Error **errp)
{
    qmp_hv_vss_cmd(false, NULL, errp);
}

static int hv_vss_pre_save(void *opaque)
{
    HvVss *vss = opaque;

    if (vss->status != HV_VSS_STATUS_IDLE) {
        if (vss->command == HV_VSS_COMMAND_FREEZE) {
            return 0;
        }
    }

    /* Don't save list of drives if `freeze' command is not being processed */
    vss->freeze_drive_count = 0;
    return 0;
}

static const VMStateDescription hv_vss_vmstate_freeze_drive = {
    .name = TYPE_HV_VSS "/freeze-drive",
    .version_id = 0,
    .minimum_version_id = 0,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32(type, struct vss_freeze_drive),
        VMSTATE_BUFFER(device_guid, struct vss_freeze_drive),
        VMSTATE_UINT8(channel, struct vss_freeze_drive),
        VMSTATE_UINT8(id, struct vss_freeze_drive),
        VMSTATE_UINT16(lun, struct vss_freeze_drive),
        VMSTATE_END_OF_LIST()
    }
};

static const VMStateDescription hv_vss_vmstate = {
    .name = TYPE_HV_VSS,
    .version_id = 0,
    .minimum_version_id = 0,
    .pre_save = hv_vss_pre_save,
    .fields = (VMStateField[]) {
        VMSTATE_STRUCT(parent, HvVss, 0, vmstate_vmbus_dev, VMBusDevice),
        VMSTATE_UINT32(command, HvVss),
        VMSTATE_UINT32(status, HvVss),
        VMSTATE_UINT32(result, HvVss),
        VMSTATE_BOOL(thaw_on_freeze, HvVss),
        VMSTATE_BOOL(freeze_hot_backup, HvVss),
        VMSTATE_UINT32(freeze_drive_count, HvVss),
        VMSTATE_STRUCT_VARRAY_UINT32(freeze_drive, HvVss, freeze_drive_count, 0,
                                     hv_vss_vmstate_freeze_drive,
                                     struct vss_freeze_drive),
        VMSTATE_UINT8(transaction, HvVss),
        VMSTATE_UINT16(version.major, HvVss),
        VMSTATE_UINT16(version.minor, HvVss),
        VMSTATE_END_OF_LIST()
    }
};

static void hv_vss_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);
    VMBusDeviceClass *vdc = VMBUS_DEVICE_CLASS(klass);

    dc->vmsd = &hv_vss_vmstate;
    qemu_uuid_parse(HV_VSS_GUID, &vdc->classid);
    qemu_uuid_parse(HV_VSS_INSTANCEID, &vdc->instanceid);
    set_bit(DEVICE_CATEGORY_MISC, dc->categories);
    vdc->channel_flags = VMBUS_CHANNEL_ENUMERATE_DEVICE_INTERFACE |
        VMBUS_CHANNEL_NAMED_PIPE_MODE;
    vdc->vmdev_realize = hv_vss_vmdev_realize;
    vdc->vmdev_reset = hv_vss_vmdev_reset;
    vdc->chan_notify_cb = hv_vss_notify_cb;
}

static const TypeInfo hv_vss_type_info = {
    .name = TYPE_HV_VSS,
    .parent = TYPE_VMBUS_DEVICE,
    .instance_size = sizeof(HvVss),
    .class_init = hv_vss_class_init,
};

static void hv_vss_register_types(void)
{
    type_register_static(&hv_vss_type_info);
}

type_init(hv_vss_register_types)
